//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is IqBal.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2005 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------


unit IqBal;

interface

uses
  SysUtils, Classes, Math, SndTypes, ComplFft, BlkSinc, Svd, Median, Dialogs,
  FileCtrl, RefMgr;

const
  REBUILD_INTERVAL: TDateTime = 1/86400;
  FILTER_SIZE = 256;
  HALF_FILTER_SIZE = FILTER_SIZE div 2;

  MIN_PRECISION = 1;
  MAX_PRECISION = 1e9;
  MAX_ENERGY = 1e10;

  MIN_SNR_SQ = 1000; //30 dB

type
  TBalancerState = (bsNoData, bsRough, bsOk);

  TIQBalancer = class
  private
    Mlf: TMlsFitter;
    LastRebuild: TDateTime;
    FFftSize: integer;
    FftMid: integer;
    //sync detector output
    MeanZ: array[0..FILTER_SIZE-1] of TComplex;
    PrecZ: array[0..FILTER_SIZE-1] of Single;
    FSamplingRate: integer;
    FBandName: string;

    procedure CollectBalanceInfo(ASpectrum: TComplexArray);
    function RebuildFilter: boolean;
    procedure CorrectSpectrum(ASpectrum: TComplexArray);
    procedure SetFftSize(const Value: integer);
    function PolyX(i: integer): Single;
    function FitPolynomial: boolean;
    procedure Leak;
    function DefaultFileName: string;
    procedure SetSamplingRate(const Value: integer);
    procedure SetBandName(Value: string);
  public
    EstimationEnabled: boolean;
    CorrectionEnabled: boolean;
    State: TBalancerState;
    GainPoints, PhasePoints: TDataPoints;
    CoeffG, CoeffF: TSingleArray;
    Quads: array[0..3] of boolean;
    Changed: boolean;
    Filter: TComplexArray;
    AutoFile: boolean;

    constructor Create;
    destructor Destroy; override;
    procedure EraseData;
    procedure Process(ASpectrum: TComplexArray);
    procedure SaveToFile(AFileName: string = '');
    procedure LoadFromFile(AFileName: string = '');

    property FftSize: integer read FFftSize write SetFftSize;
    property SamplingRate: integer read FSamplingRate write SetSamplingRate;
    property BandName: string read FBandName write SetBandName;
  end;



implementation

{ TIQBalancer }

constructor TIQBalancer.Create;
begin
  Mlf := TMlsFitter.Create;
  Mlf.Tolerance := 1E-4;

  EstimationEnabled := true;
  CorrectionEnabled := true;
end;


destructor TIQBalancer.Destroy;
begin
  if AutoFile then SaveToFile;
  Mlf.Free;
  inherited;
end;


procedure TIQBalancer.EraseData;
begin
  Filter := nil;
  FillChar(MeanZ, SizeOf(MeanZ), 0);
  FillChar(PrecZ, SizeOf(PrecZ), 0);
  State := bsNoData;
  GainPoints := nil;
  PhasePoints := nil;
  CoeffG := nil;
  CoeffF := nil;
  FillChar(Quads, SizeOf(Quads), 0);
  Changed := true;
end;


procedure TIQBalancer.SetFftSize(const Value: integer);
begin
  if FFftSize = Value then Exit;
  FFftSize := Value;
  FftMid := Value div 2;
  RebuildFilter;
end;


procedure TIQBalancer.Process(ASpectrum: TComplexArray);
begin
  Changed := false;

  //analyze input data and update correction filter
  if EstimationEnabled then
    begin
    CollectBalanceInfo(ASpectrum);

    if (Now - LastRebuild) > REBUILD_INTERVAL then
      begin
      Changed := RebuildFilter;
      Leak;
      end;
    end;

  //apply correction
  if CorrectionEnabled and (Filter <> nil)
    then CorrectSpectrum(ASpectrum);
end;


//map index (0..filtersize) to -0.5..0.5
//the negative part is in the upper half of the array
function TIQBalancer.PolyX(i: integer): Single;
begin
  Result := i / FILTER_SIZE;
  if Result > 0.5 then Result := Result - 1;
end;


function TIQBalancer.RebuildFilter: boolean;
var
  i, p: integer;
  x, Gain, Phase: Single;
begin
  LastRebuild := Now;
  Result := false;

  //calculate CoeffG and CoeffF
  if not FitPolynomial then Exit;

  //generate frequency response of the ideal correction filter
  Filter := nil; SetLength(Filter, FftSize);
  for i:=0 to FILTER_SIZE-1 do
    begin
    x := PolyX(i);
    Gain := CoeffG[0];
    for p:=1 to High(CoeffG) do Gain := Gain + CoeffG[p]*IntPower(x,p);
    Phase := CoeffF[0];
    for p:=1 to High(CoeffF) do Phase := Phase + CoeffF[p]*IntPower(x,p);

    Filter[i].Re := Gain * Cos(Phase) / (FILTER_SIZE);
    Filter[i].Im := Gain * Sin(Phase) / (FILTER_SIZE);
    end;

  //transform to the time domain
  SetLength(Filter, FILTER_SIZE);
  ComplexIFFT(Filter);
  SetLength(Filter, FftSize);

  //apply window function
  //note: FILTER_SIZE << Length(Filter), the kernel is zero-padded
  for i:=0 to FILTER_SIZE-1 do
    begin
    x := 0.5 + PolyX(i);
    Filter[i].Re := Filter[i].Re * BlackmanWin(x);
    Filter[i].Im := Filter[i].Im * BlackmanWin(x);
    end;

  //back to the frequency domain
  //the frequency response is now interpolated over the whole FftSize
  Move(Filter[HALF_FILTER_SIZE], Filter[FftSize-HALF_FILTER_SIZE], HALF_FILTER_SIZE * SizeOf(TComplex));
  FillChar(Filter[HALF_FILTER_SIZE], HALF_FILTER_SIZE * SizeOf(TComplex), 0);
  ComplexFFT(Filter);

  Result := true;
end;


procedure TIQBalancer.CorrectSpectrum(ASpectrum: TComplexArray);
var
  Src: TComplexArray;
  i, j: integer;
  Sp, Sm, Err: TComplex;
begin
  if Filter = nil then Exit;

  Assert(Length(ASpectrum) = FftSize);

  Src := Copy(ASpectrum);

  //apply filter to the Q channel
  for i:=0 to FftSize-1 do
    begin
    j := (FftSize-i) mod FftSize; //mod helps when i = 0
    Sp := Src[i];
    Sm := Src[j];
    Err :=  Filter[j];

    ASpectrum[i].Re := Sp.Re + Sm.Re - (Sp.Im + Sm.Im) * Err.Im + (Sp.Re - Sm.Re) * Err.Re;
    ASpectrum[i].Im := Sp.Im - Sm.Im + (Sp.Im + Sm.Im) * Err.Re + (Sp.Re - Sm.Re) * Err.Im;
    end;
end;


function TIQBalancer.FitPolynomial: boolean;
var
  i, j, Dst: integer;
  Qg, Qf: array[0..3] of integer;
  Den: Single;
  Z: TComplex;
  Gain, Phase: Single;
  ResidG, ResidF: TSingleArray;
  IMdn: Single;

  W: array[0..3] of Single;
begin
  Result := false;
  State := bsNoData;

  //collect reliable points from the synchronous detector output
  FillChar(Qg, SizeOf(Qg), 0);
  SetLength(GainPoints, FILTER_SIZE);
  SetLength(PhasePoints, FILTER_SIZE);
  Dst := 0;
  for i:=0 to FILTER_SIZE-1 do
    begin
    //reject inaccurate estimates
    if PrecZ[i] < MIN_PRECISION then Continue;

    //reject wild estimates
    try
      Z := MeanZ[i];
      if Abs(Z.Re) >= 0.5 then Continue;
      Den := Sqrt(1 - Sqr(2*Z.Re));
      Phase := ArcSin(2*Z.Im / Den);
      Gain := Den / (1 - 2*Z.Re);
      if (Abs(Phase*180/Pi) > 20) or (Abs(Ln(Gain + 1E-30)) > 1) then Continue;

      //put estimates in the arrays
      GainPoints[Dst].x := PolyX(i);
      GainPoints[Dst].y := Gain;
      GainPoints[Dst].w :=  PrecZ[i];

      PhasePoints[Dst].x := PolyX(i);
      PhasePoints[Dst].y := Phase;
      PhasePoints[Dst].w := PrecZ[i];

      //remember which quarters of passband are covered
      Qg[Trunc(2 + 4*GainPoints[Dst].x)] := 1;

      //count good estimates
      Inc(Dst);
    except Continue; end;

    end;

  Move(Qg, Qf, SizeOf(Qg));
  SetLength(GainPoints, Dst);
  SetLength(PhasePoints, Dst);
  if Dst = 0 then Exit;


  //Q-weights
  FillChar(W, SizeOf(W), 0);
  for i:=0 to Dst-1 do
    begin
    j := Trunc(2 + 4*GainPoints[i].x);
    W[j] := W[j] + GainPoints[i].w;
    end;
  for i:=0 to Dst-1 do
    begin
    j := Trunc(2 + 4*GainPoints[i].x);
    GainPoints[i].w := GainPoints[i].w / {Sqrt}(W[j]);
    PhasePoints[i].w := GainPoints[i].w;
    end;
  Move(Qf, Qg, SizeOf(Qg));


  //weighted MLS polynomial fitting.
  //if data in all 4 quarters are available, fit 3-rd order polynomials,
  //otherwise use lower orders

  //fit polynomial to gain estimates
  Mlf.Data := GainPoints;
  case Qg[0]+Qg[1]+Qg[2]+Qg[3] of
    1,2: Mlf.Fit(1);
    3:   Mlf.Fit(3);
    4:   Mlf.Fit(4);
    end;


  //robust fitting, bi-square method
  //http://www.stat.sfu.ca/~weldon/400-030915-ch3.pdf
  if (Qg[0]+Qg[1]+Qg[2]+Qg[3]) = 4 then
    begin
    //calculate normalized residuals
    SetLength(ResidG, Length(GainPoints));
    for i:=0 to High(ResidG) do ResidG[i] := Abs(Mlf.CalcResidual(i));
    IMdn := 1 / (6 * CalcMedian(ResidG) + 1e-10);
    for i:=0 to High(ResidG) do ResidG[i] := ResidG[i] * IMdn;
    FillChar(Qg, SizeOf(Qg), 0);
    Dst := 0;
    //downweight or eliminate outliers
    for i:=0 to High(GainPoints) do
      if ResidG[i] < 1 then
        begin
        GainPoints[Dst] := GainPoints[i];
        GainPoints[Dst].w :=  GainPoints[Dst].w * Sqr(1 - Sqr(ResidG[i]));
        Inc(Dst);
        Qg[Trunc(2 + 4*GainPoints[i].x)] := 1;
        end;
    SetLength(GainPoints, Dst);
    //fit again
    Mlf.Data := GainPoints;
    case Qg[0]+Qg[1]+Qg[2]+Qg[3] of
      1,2: Mlf.Fit(1);
      3:   Mlf.Fit(3);
      4:   Mlf.Fit(4);
      end;
    end;

  CoeffG := Copy(Mlf.Coeff);



  //same algirithm for phase

  Mlf.Data := PhasePoints;
  case Qf[0]+Qf[1]+Qf[2]+Qf[3] of
    1:   Mlf.Fit(1);
    2,3: Mlf.Fit(2);
    4:   Mlf.Fit(4);
    end;

  if (Qf[0]+Qf[1]+Qf[2]+Qf[3]) = 4 then
    begin
    SetLength(ResidF, Length(PhasePoints));
    for i:=0 to High(ResidF) do ResidF[i] := Abs(Mlf.CalcResidual(i));
    IMdn := 1 / (6 * CalcMedian(ResidF));
    for i:=0 to High(ResidF) do ResidF[i] := ResidF[i] * IMdn;
    FillChar(Qf, SizeOf(Qf), 0);
    Dst := 0;
    for i:=0 to High(PhasePoints) do
      if ResidF[i] < 1 then
        begin
        PhasePoints[Dst] := PhasePoints[i];
        PhasePoints[Dst].w :=  PhasePoints[Dst].w * Sqr(1 - Sqr(ResidF[i]));
        Inc(Dst);
        Qf[Trunc(2 + 4*PhasePoints[i].x)] := 1;
        end;
    SetLength(PhasePoints, Dst);
    Mlf.Data := PhasePoints;
    case Qf[0]+Qf[1]+Qf[2]+Qf[3] of
      1:   Mlf.Fit(1);
      2,3: Mlf.Fit(2);
      4:   Mlf.Fit(4);
      end;
    end;

  CoeffF := Copy(Mlf.Coeff);


  //sanity check

  if (Length(CoeffG) = 4) and (Abs(CoeffG[3]) > 0.3) then
    begin
    Mlf.Data := GainPoints;
    Mlf.Fit(3);
    CoeffG := Copy(Mlf.Coeff);
    end;

  if (Length(CoeffF) = 4) and (Abs(CoeffF[3]) > 0.3) then
    begin
    Mlf.Data := PhasePoints;
    Mlf.Fit(2);
    CoeffF := Copy(Mlf.Coeff);
    end;



  //mark covered quarters, update state

  for i:=0 to 3 do
    begin
    Qg[i] := Qg[i] * Qf[i];
    Quads[i] := Qg[i] = 1;
    end;

  case Qg[0] + Qg[1] + Qg[2] + Qg[3] of
    0:    State := bsNoData;
    1..3: State := bsRough;
    4:    State := bsOk;
    end;

  Result := true;
end;


procedure TIQBalancer.CollectBalanceInfo(ASpectrum: TComplexArray);
const
  LEAK: Double = 1E11;
var
  SnrSqr: TSingleArray;
  i, j: integer;
  Len, Mid, InMargin, OutMargin: integer;
  Shrink: integer;
  Z: TComplex;
  Pwr: Single;
  IMdn: Single;
  Precision: Single;
begin
  //the input spectrum is oversampled due to windowing, decimate
  Len := FftSize div 2;

  Shrink := Len div FILTER_SIZE;
  Mid := Len div 2;

  InMargin := Round(5000{Hz} * Len / SamplingRate);
  if SamplingRate = 96000
    then OutMargin := Round(5000 * Len / SamplingRate)
    else OutMargin := Round(2000 * Len / SamplingRate);

  //signal-to-noise ratio squared
  SetLength(SnrSqr, Len);
  for i:=0 to High(SnrSqr) do
    SnrSqr[i] := Sqr(ASpectrum[i*2].Re) + Sqr(ASpectrum[i*2].Im);
  IMdn := 1 / (CalcMedian(SnrSqr) +1E-30);
  for i:=0 to High(SnrSqr) do SnrSqr[i] := SnrSqr[i] * IMdn;

  //find peak
  for i:=InMargin to Len-InMargin do
    if (Abs(i-Mid) > OutMargin) and
       //check snr of the main signal, main/image ratio, and surrounding values
       (SnrSqr[i] > MIN_SNR_SQ) and (SnrSqr[Len-i] > 0) and ((SnrSqr[i] / SnrSqr[Len-i]) > 10)  and
       (SnrSqr[i] > SnrSqr[i+1]) and (SnrSqr[i] > SnrSqr[i-1]) and
       (SnrSqr[i] > SnrSqr[i+2]) and (SnrSqr[i] > SnrSqr[i-2]) and
       (SnrSqr[i] > SnrSqr[i+3]) and (SnrSqr[i] > SnrSqr[i-3]) then
      begin
      //synchronous detector
      j := i * 2;
      Z := ComplexMul(ASpectrum[j], ASpectrum[FftSize-j]);
      Pwr := Sqr(ASpectrum[j].Re) + Sqr(ASpectrum[j].Im) +
             Sqr(ASpectrum[FftSize-j].Re) + Sqr(ASpectrum[FftSize-j].Im);
      Z.Re := Z.Re / Pwr;
      Z.Im := Z.Im / Pwr;

      //precision of observation
      Precision := 1E-6 * (SnrSqr[i] + SnrSqr[Len-i]);
      if Precision > 1E10 then Continue;

      //group peaks
      j := Round(i / Shrink);

      //precision of estimate
      PrecZ[j] := PrecZ[j] + Precision;
      PrecZ[j] := Min(PrecZ[j], MAX_PRECISION);
      if PrecZ[j] = 0 then Continue;

      //weighted mean
      try
        Z.Re := (MeanZ[j].Re * (PrecZ[j] - Precision) + Z.Re * Precision) / PrecZ[j];
        Z.Im := (MeanZ[j].Im * (PrecZ[j] - Precision) + Z.Im * Precision) / PrecZ[j];
        MeanZ[j] := Z;
      except end;
      end;
end;


procedure TIQBalancer.Leak;
var
  i: integer;
  Energy: Double;
begin
  Energy := 0;
  for i:=0 to High(PrecZ) do Energy := Energy + PrecZ[i];
  if Energy < MAX_ENERGY then Exit;
  for i:=0 to High(PrecZ) do PrecZ[i] := PrecZ[i] * MAX_ENERGY / Energy;
end;


function TIQBalancer.DefaultFileName: string;
var
  Dir: string;
begin
  Dir := ExtractFilePath(GetIniName) + 'Bal\';
  ForceDirectories(Dir);

  if FBandName = ''
    then Result := Dir + Format('IQBalance%d.dat', [SamplingRate div 1000])
    else Result := Dir + Format('IQBalance%d-%s.dat', [SamplingRate div 1000, FBandName]);
end;


procedure TIQBalancer.LoadFromFile(AFileName: string);
begin
  EraseData;

  if AFileName = '' then AFileName := DefaultFileName;
  if not FileExists(AFileName) then Exit;

  with TFileStream.Create(AFileName, fmOpenRead) do
    try
      ReadBuffer(MeanZ, SizeOf(MeanZ));
      ReadBuffer(PrecZ, SizeOf(PrecZ));
    finally
      Free;
    end;

  RebuildFilter;
end;


procedure TIQBalancer.SaveToFile(AFileName: string);
begin
  if FSamplingRate = 0 then Exit;

  if AFileName = '' then AFileName := DefaultFileName;

  with TFileStream.Create(AFileName, fmCreate) do
    try
      WriteBuffer(MeanZ, SizeOf(MeanZ));
      WriteBuffer(PrecZ, SizeOf(PrecZ));
    finally
      Free;
    end;
end;


procedure TIQBalancer.SetSamplingRate(const Value: integer);
begin
  if Value = FSamplingRate then Exit;
  if AutoFile then SaveToFile;
  FSamplingRate := Value;
  if AutoFile then LoadFromFile;
end;


procedure TIQBalancer.SetBandName(Value: string);
var
  i: integer;
begin
  Value := Trim(Value);
  for i:=Length(Value) downto 1 do
    if Value[i] = '&' then Delete(Value, i, 1)
    else if not (Value[i] in ['A'..'Z', 'a'..'z','0'..'9']) then Value[i] := '_';

  if FBandName = Value then Exit;
  if AutoFile then SaveToFile;
  FBandName := Value;
  if AutoFile then LoadFromFile;
end;



end.